from typing import Iterable

import polars as pl


def aggregate_cod(
    cause_specific_rates: pl.LazyFrame,
    by: Iterable[str],
    other_exprs: Iterable[pl.Expr] = [],
) -> pl.LazyFrame:
    """
    Aggregates age-specific COD across causes of death and
    cancer sites within a particular cell.
    """
    return cause_specific_rates.group_by(*by, "age", maintain_order=True).agg(
        pl.col("mortality").sum(), *other_exprs
    )



"""
See e.g. 
- https://www.ssa.gov/oact/HistEst/PerLifeTables/LifeTableDefinitions.pdf
- https://www.aitrs.org/sites/default/files/Life%20Tables.pdf
- https://mortality.org/File/GetDocument/Public/Docs/MethodsProtocolV6.pdf (p. 36)

Expects columns "mortality" and "age", with the data frame sorted (increasing)
by age, for example
```python
┌─────┬───────────┐
│ age ┆ mortality │
│ --- ┆ ---       │
│ i64 ┆ f64       │
╞═════╪═══════════╡
│ 30  ┆ 0.00099   │
│ 31  ┆ 0.001018  │
│ 32  ┆ 0.001047  │
│ 33  ┆ 0.001076  │
│ 34  ┆ 0.001104  │
│ …   ┆ …         │
│ 80  ┆ 0.000004  │
│ 81  ┆ 0.000004  │
│ 82  ┆ 0.000004  │
│ 83  ┆ 0.000005  │
│ 84  ┆ 0.000005  │
└─────┴───────────┘
```
"""

from typing import Iterable
import polars as pl
from polars._typing import IntoExpr


def create_life_table(
    raw_mortality_rates: pl.LazyFrame,  # *by, age, raw_mortality_rate
    *,
    by: Iterable[str] = [],
    m: pl.Expr = pl.col("mortality"),
    age: pl.Expr = pl.col("age"),
    initial_cohort_size: IntoExpr = 1,
    separation_factor: IntoExpr = 0.5,
    infant_separation_factor: IntoExpr = 0.14,
    final_separation_factor: IntoExpr = 1 / pl.col("mortality"),
    q_equals_m=False,
) -> pl.LazyFrame:

    # TODO: handle age bins of width other than 1
    # width of age bin (next age - current age)
    # width = age.shift(-1) - age

    is_final_age = age == age.max()

    # mean number of years lived for those within bin (separation factor)
    s = (
        # infants
        pl.when(age=0)
        .then(infant_separation_factor)
        # TODO: handle final age bin
        .when(is_final_age)
        .then(final_separation_factor)
        # otherwise 1/2
        .otherwise(separation_factor)
    )

    # use q != m if populations are measured mid-year (after some number of people have died)
    q_inner = m if q_equals_m else m / (1 + (1 - s) * m)

    # probability of dying at age x, conditional on having lived until age x
    # top age bin gets set to 1
    q = pl.when(is_final_age).then(1.0).otherwise(q_inner)

    # probability of living to age x
    l = (1 - q).shift(1, fill_value=initial_cohort_size).cum_prod()

    # probability of dying at age x
    d = l * q

    # person years lived at exact age
    L = l - (1 - s) * d

    # person years remaining
    T = L.cum_sum(reverse=True)

    # life expectancy remaining
    e = T / l

    life_table_columns = dict(q=q, l=l, d=d, s=s, L=L, T=T, e=e)

    return raw_mortality_rates.sort(*by, "age").with_columns(
        expr.over(by).alias(name) if by else expr.alias(name)
        for name, expr in life_table_columns.items()
    )


def compute_le(
    mortality_rates: pl.LazyFrame,
    *,
    by: Iterable[str] = [],
    m: pl.Expr = pl.col("mortality"),
    age: pl.Expr = pl.col("age"),
    **kwargs,
) -> pl.LazyFrame:
    return create_life_table(mortality_rates, by=by, age=age, m=m, **kwargs).select(
        *by, age, le=pl.col("e") + age
    )



"""
original paper: https://doi.org/10.2307/2061029
more concrete implementation: https://doi.org/10.1016/j.annepidem.2014.05.006
"""

from typing import Iterable, Literal, Optional

import polars as pl
from polars import selectors as cs

from lifetables.helpers import aggregate_cod
from lifetables.life_table import create_life_table

_SHARE_WITHIN_AGE = "share_within_age"


def _arriaga_decomposition_by_age(
    initial_mortality_rates: pl.LazyFrame,
    new_mortality_rates: pl.LazyFrame,
    *,
    by: Iterable[str] = [],
):

    q, l, L = pl.col("mortality"), pl.col("l"), pl.col("L")
    q_new, l_new, L_new = pl.col("mortality_new"), pl.col("l_new"), pl.col("L_new")

    e_new = pl.col("e_new")
    T_new = pl.col("T_new")

    direct_effect = (l * (L_new / l_new - L / l)).alias("direct_effect")
    indirect_effect = (T_new.shift(-1) * (l / l_new).diff(-1)).alias("indirect_effect")

    # gap contributions
    years = (direct_effect + indirect_effect).alias("contribution_years")
    proportion = (years / years.sum()).alias("contribution_proportion")

    return (
        initial_mortality_rates.pipe(create_life_table, by=by)
        .join(
            new_mortality_rates.pipe(create_life_table, by=by),
            on=list(by) + ["age"],
            how="inner",
            validate="1:1",
            suffix="_new",
        )
        .select(
            *by,
            "age",
            # direct_effect.over(by),
            # indirect_effect.over(by),
            years.over(by),
            proportion.over(by),
        )
    )


def _arriaga_decomposition_by_cause_within_age(
    initial_mortality_rates: pl.LazyFrame,
    new_mortality_rates: pl.LazyFrame,
    *,
    by: Iterable[str],
    cause_column="cause_of_death",
):
    """
    returns df w/ columns `*by, age, cause_column, share_within_age`
    """

    q = pl.col("mortality")
    q_new = pl.col("mortality_new")
    cause_share = (q_new - q) / (q_new - q).sum().over(*by, "age")

    return initial_mortality_rates.join(
        new_mortality_rates,
        on=list(by) + ["age", cause_column],
        how="inner",
        validate="1:1",
        suffix="_new",
    ).select(*by, "age", cause_column, cause_share.alias(_SHARE_WITHIN_AGE))


def _arriaga_decomposition(
    initial_mortality_rates: pl.LazyFrame,
    new_mortality_rates: pl.LazyFrame,
    *,
    by: Iterable[str] = [],
    within_age: Optional[str] = None,  # e.g. cause_of_death
):
    """
    returns df w/ columns `*by, age, cause_column?, contribution_years, contribution_proportion`
    """

    # calculate contributions to gap by age
    contributions_by_age = _arriaga_decomposition_by_age(
        initial_mortality_rates.pipe(aggregate_cod, by=by),
        new_mortality_rates.pipe(aggregate_cod, by=by),
        by=by,
    )

    if not within_age:
        return contributions_by_age

    # calculate contribution shares within age
    shares_within_age = _arriaga_decomposition_by_cause_within_age(
        initial_mortality_rates,
        new_mortality_rates,
        by=by,
        cause_column=within_age,
    )

    # multiply total age contribution by within-age share
    # to get total contribution by cause x age
    return (
        shares_within_age.join(
            contributions_by_age,
            on=list(by) + ["age"],
            how="inner",
            validate="m:1",
        )
        .with_columns(cs.starts_with("contribution_") * pl.col(_SHARE_WITHIN_AGE))
        .drop(_SHARE_WITHIN_AGE)
    )


def arriaga_decomposition(
    initial_mortality: pl.LazyFrame,
    new_mortality: Optional[pl.LazyFrame] = None,
    *,
    by: Iterable[str] = [],
    direction: Literal["forward", "backward", "average"] = "average",
    within_age: Optional[str] = None,  # e.g. cause_of_death
) -> pl.LazyFrame:

    # if other mortality rates are not passed, set them uniformly to zero
    # to decompose the le toll by age/cause
    if new_mortality is None:
        new_mortality = initial_mortality.with_columns(pl.lit(0).alias("mortality"))

    if direction == "forward":
        return _arriaga_decomposition(
            initial_mortality,
            new_mortality,
            by=by,
            within_age=within_age,
        )

    if direction == "backward":
        return _arriaga_decomposition(
            new_mortality,  # flip initial and new
            initial_mortality,
            by=by,
            within_age=within_age,
        ).with_columns(pl.col("contribution_years").mul(-1)) # correct negative sign in years

    # now direction == "average":

    directions = ["forward", "backward"]

    both_directions = pl.concat(
        arriaga_decomposition(
            initial_mortality,
            new_mortality,
            by=by,
            direction=d,  # type: ignore
            within_age=within_age,
        ).with_columns(pl.lit(d).alias("direction"))
        for d in directions
    )

    group = [*by, "age", within_age] if within_age else [*by, "age"]
    return both_directions.group_by(group, maintain_order=True).agg(
        cs.exclude("direction").mean()
    )



import pandas as pd
import polars as pl


import polars.selectors as cs


def get_standard_pops(
    url="https://seer.cancer.gov/stdpopulations/stdpop.singleages.html",
    age_column=cs.by_index(0),
    pop_column=cs.by_index(1),
):
    raw = pl.from_pandas(pd.read_html(url).pop())

    return raw.select(
        age_column.str.extract(r"(\d{2}) years")
        .str.to_integer(strict=False)
        .alias("age"),
        pop_column.alias("population"),
    ).drop_nulls()



from typing import Optional, Literal
from importlib import resources

import polars as pl
from polars import selectors as cs


DATA = resources.files("lifetables.data")
SEX_TYPE = pl.Enum(("Male", "Female"))


def _hmd_life_table(
    sex: Literal["Male", "Female", "Both"],
) -> pl.DataFrame:

    path = DATA / "hmd" / f"{sex[0].lower()}ltper_1x1.txt"

    lines = [l.split() for l in path.read_text().splitlines()[2:]]

    return pl.from_records(lines[1:], orient="row", schema=lines[0]).select(
        pl.col("Year", "Age")
        .str.extract(r"(\d+)")
        .str.to_integer()
        .name.to_lowercase(),
        cs.ends_with("x").cast(pl.Float64).name.map(lambda s: s.strip("x")),
    )


def hmd_life_table(by_sex=True) -> pl.DataFrame:
    """
    From the Human Mortality Database:
    https://www.mortality.org/Country/Country?cntr=USA
    """
    if by_sex:
        return pl.concat(
            _hmd_life_table(sex).select(
                pl.lit(sex).cast(SEX_TYPE).alias("sex"), pl.all()
            )
            for sex in ("Male", "Female")
        )

    return _hmd_life_table("Both")



import polars as pl

from lifetables.life_table import create_life_table
from lifetables.data import hmd_life_table


def test_life_table() -> None:
    """
    Test to check whether we match the life expectancy estimates
    of the Human Mortality Database when using their mortality rates.
    """

    # original life table from HMD
    original = hmd_life_table().sort(pl.all())

    # recreation using our life table function
    recreation = (
        original.lazy()
        .select("year", "age", "sex", "m", "a")
        .pipe(
            create_life_table,
            by=["year", "sex"],
            m=pl.col("m"),
            final_separation_factor=1 / pl.col("m"),
            infant_separation_factor=pl.col("a"),
            initial_cohort_size=100_000,
        )
        .sort(pl.all())
        .drop("a")
        .collect()
    )

    # compare LE estimates
    error = pl.col("e").sub("e_right")
    mse, mean, max = (
        original.join(recreation, on=["year", "sex", "age"])
        .select(
            mse=error.pow(2).mean(),
            mean=error.mean(),
            max=error.abs().max(),
        )
        .iter_rows()
        .__next__()
    )

    assert max < 0.01
    assert mse < 1e-5
    assert abs(mean) < 1e-5



from typing import Iterable, Optional, Sequence

import polars as pl
from polars import selectors as cs

from lifetables.populations import get_standard_pops


def smooth_mortality_rates():
    raise NotImplementedError


def align_mortality_rates(
    reference_rates: pl.LazyFrame,
    rates_to_align: pl.LazyFrame,
    *,
    by: Iterable[str],
    age_to_align: pl.Expr = pl.col("age").max(),  # executed on reference_rates
    column_to_align="mortality",
):
    # *by, shift
    shifts = rates_to_align.join(
        reference_rates.filter(pl.col("age").eq(age_to_align).over(by)),
        how="inner",
        validate="1:m",
        on=[*by, "age"],
    ).select(
        *by,
        shift=pl.col(column_to_align + "_right") / pl.col(column_to_align),
    )

    return rates_to_align.join(
        shifts,
        on=[*by],
        how="inner",
        validate="m:1",
    ).select(*by, "age", pl.col(column_to_align) * pl.col("shift"))


def fill_mortality_rates(
    original_rates: pl.LazyFrame,
    rates_to_fill_with: pl.LazyFrame,
    *,
    by: Iterable[str],  # e.g. ["year", "sex", "race", "cause_of_death"]
    join_on: Iterable[str],  # e.g. ["year", "sex"] or ["year"]
    min_age=30,
    switch_age=85,
    max_age=110,
    pre_switch_fill_value=float("nan"),
    post_switch_fill_value=float("nan"),
    column_to_fill="mortality",
) -> pl.LazyFrame:
    # construct spine containing the full age range of interest
    ages = pl.int_range(min_age, max_age + 1, step=1, eager=True).to_frame("age").lazy()
    spine = original_rates.select(*by).unique().sort(pl.all()).join(ages, how="cross")

    # populate spine with both original rates and fill rates
    both_rates = spine.join(
        original_rates,
        how="left",
        on=list(by) + ["age"],
        validate="1:1",
    ).join(
        rates_to_fill_with.select(*join_on, "age", column_to_fill),
        how="left",
        on=list(join_on) + ["age"],
        validate="m:1",
        suffix="_fill",
    )

    return (
        both_rates.with_columns(
            # before swich_age, use original rates
            pl.when(pl.col("age").lt(switch_age))
            .then(pl.col(column_to_fill).fill_null(value=pre_switch_fill_value))
            # on and after swich_age, use new rates
            .otherwise(
                pl.col(column_to_fill + "_fill").fill_null(value=post_switch_fill_value)
            )
        )
        # drop the fill rates
        .drop(cs.ends_with("_fill"))
        .sort(*by, "age")
    )


def age_standardized_mortality(
    mortality_rates: pl.LazyFrame,
    *,
    by: Iterable[str],
    standard_populations: Optional[pl.LazyFrame] = None, # age, population
    mortatliy_col=pl.col("mortality"),
    other_exprs: Iterable[pl.Expr] = [],
) -> pl.LazyFrame:
    standard_populations = standard_populations or get_standard_pops().lazy()

    return (
        mortality_rates.join(
            standard_populations.select(
                "age",
                w=pl.col("population") / pl.col("population").sum(),
            ),
            how="left",
            validate="m:1",
            on=["age"],
        )
        .group_by(*by)
        .agg(mortatliy_col.dot("w"), *other_exprs)
    )
